local MAJOR = "LibScriptableDisplayError-1.0"
local MINOR = 15
assert(LibStub, MAJOR.." requires LibStub")
local LibScriptableDisplayError = LibStub:NewLibrary(MAJOR, MINOR)
if not LibScriptableDisplayError then return end
local pool = setmetatable({}, {__mode = "k"})
LibScriptableDisplayError.defaultTexts = {"Quiet",  "Normal", "Verbose"}

local DEFAULT_LEVEL = #LibScriptableDisplayError.defaultTexts
-- 3 verbose
-- 2 warning
-- 1 quiet

if not LibScriptableDisplayError.__index then
	LibScriptableDisplayError.__index = LibScriptableDisplayError
	LibScriptableDisplayError.name = MAJOR
	LibScriptableDisplayError.errorLevel = DEFAULT_LEVEL
end

-- Create a new LibScriptableDisplayError object
-- @name LibScriptableDisplayError:New
-- @usage LibScriptableDisplayError:New(name, [handler])
-- @param name Give your error object a name
-- @param errorLevel Critical level for this object. This is optional. Default is 2.
-- @param handler An optional handler function, otherwise we use ChatFrame1 for printing
-- @return A new LibScriptableDisplayError object
function LibScriptableDisplayError:New(name, errorLevel, handler)
	assert(name, "LibScriptableDisplayError.New requires a name")
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)

	obj.name = name
	obj.errorLevel = errorLevel or DEFAULT_LEVEL

	if type(handler) == "function" then
		obj.handler = handler
	end

	return obj
end
-- Delete a LibScriptableDisplayError object
-- @name LibScriptableDisplayError:Del
-- @usage object:Del() or LibScriptableDisplayError(object)
-- @return Nothing
function LibScriptableDisplayError:Del(err)
	if not err then err = self end
	pool[err] = true
end

-- Print something
-- @name LibScriptableDisplayError.Print
-- @usage LibError:Print(text, errorLevel)
-- @param text The text to print.
-- @param errorLevel The critical threshhold for this LibError object
-- @return Nothing
function LibScriptableDisplayError:Print(text, errorLevel)
	assert(self, "LibScriptableDisplayError failed printing due to invalid 'self'")
	assert(self.name and self.name ~= "", "LibScriptableDisplayError object has no name")
	assert(text, "LibScriptableDisplayError failed printing due to nil text (" .. self.name .. ")")

	if type(errorLevel) ~= "number" then
		errorLevel = DEFAULT_LEVEL
	end

	if self.handler then
		self.handler(text, errorLevel)
		return
	end

	local color
	if errorLevel == 3 then
		color = "|cffff0000"
	elseif errorLevel == 2 then
		color = "|cffffff00"
	elseif errorLevel == 1 then
		color = "|cff00ff00"
	else
		color = "|cffffffff"
	end

	if self.errorLevel >= errorLevel then
		ChatFrame1:AddMessage(("(%s)"):format(date()) .. color .. (self.name or " ") .. "|r:: " .. text)
	end
end